"""This package contains all of the Pydantic models that are used to interact
with the Anacreon API

Because all of the types are Pydantic models, they come with handy methods that
can serialize/deserialize to JSON. You can read more about Pydantic on their
`documentation website <https://pydantic-docs.helpmanual.io/>`_.
"""
